class DesktopLibrary:
    raise DeprecationWarning('This code has been moved to Robotframework-ApplicationLibrary as of version 3.0.0, more information: https://github.com/Accruent/robotframework-zoomba/releases/tag/3.0.0')



import itertools
from SeleniumLibrary import SeleniumLibrary
from robot.libraries.BuiltIn import BuiltIn
from robot.api.deco import keyword
from robot.libraries.Collections import Collections
from time import time
from robot.utils import is_string
import importlib

# Importing ReactSelect
try:
    RS = importlib.import_module('Helpers.ReactSelect', package='Helpers')
except ModuleNotFoundError:
    RS = importlib.import_module('.Helpers.ReactSelect', package='Zoomba')


zoomba = BuiltIn()
zoomba_collections = Collections()
SCREENSHOT_COUNTER = itertools.count()


class GUILibrary(SeleniumLibrary):
    """Zoomba GUI Library

    This class inherits from the SeleniumLibrary Class, and expands it with some commonly used keywords. For more information on SeleniumLibrary please visit: https://robotframework.org/SeleniumLibrary/SeleniumLibrary.html
    """

    def __init__(self, timeout=5.0, implicit_wait=0.0, run_on_failure='Save Selenium Screenshot',
                 screenshot_root_directory=None, plugins=None, event_firing_webdriver=None):
        """GUILibrary can be imported with several optional arguments.

        - ``timeout``:
          Default value for `timeouts` used with ``Wait ...`` keywords.
        - ``implicit_wait``:
          Default value for `implicit wait` used when locating elements.
        - ``run_on_failure``:
          Default action for the `run-on-failure functionality`.
        - ``screenshot_root_directory``:
          Path to folder where possible screenshots are created or EMBED.
          See `Set Screenshot Directory` keyword for further details about EMBED.
          If not given, the directory where the log file is written is used.
        - ``plugins``:
          Allows extending the GUILibrary with external Python classes.
        - ``event_firing_webdriver``:
          Class for wrapping Selenium with
          [https://seleniumhq.github.io/selenium/docs/api/py/webdriver_support/selenium.webdriver.support.event_firing_webdriver.html#module-selenium.webdriver.support.event_firing_webdriver|EventFiringWebDriver]
        """
        super().__init__(timeout, implicit_wait, run_on_failure, screenshot_root_directory, plugins,
                         event_firing_webdriver)

    @keyword("Element Value Should Be Equal")
    def element_value_should_be_equal(self, locator, expected_value):
        """Assert that the value attribute of a web element is equal to a given value\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)\n
        expected_value: (string) A string or int value to be compared against the actual web element value
        """
        value = self.get_value(locator)
        zoomba.should_be_equal(value, expected_value)

    @keyword("Element Value Should Not Be Equal")
    def element_value_should_not_be_equal(self, locator, expected_value):
        """Assert that the value attribute of a web element is not equal to a given value\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)\n
        expected_value: (string) A string or int value to be compared against the actual web element value
        """
        value = self.get_value(locator)
        zoomba.should_not_be_equal(value, expected_value)

    @keyword("Wait For And Click Element")
    def wait_for_and_click_element(self, locator, timeout=None):
        """This is a series of chained Selenium keywords, that tries to find a web element first, and then clicks it.
        If the element fails to be clicked, it will scroll to the bottom of the page and try again.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.click_element(locator)

    @keyword('Wait For And Input Text')
    def wait_for_and_input_text(self, locator, text, timeout=None):
        """This is a series of chained Selenium keywords, that tries to find a web element first, and then input text.
        If the element fails to typed into, it will scroll to the bottom of the page and try again.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)\n
        text: (string) Text to be typed into the input field.
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.input_text(locator, text)

    @keyword('Wait For And Input Password')
    def wait_for_and_input_password(self, locator, password, timeout=None):
        """This is a series of chained Selenium keywords, that tries to find a web element first, and then input text.
        The password text is not printed displayed in logs.
        If the element fails to typed into, it will scroll to the bottom of the page and try again.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)\n
        password: (string) Password text to be typed into the input field.
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.input_password(locator, password)

    @keyword("Wait For And Select Frame")
    def wait_for_and_select_frame(self, locator, timeout=None):
        """This is a series of chained Selenium keywords, that first waits until an iFrame exists in the page, and then
        selects it.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_until_page_contains_element(locator, timeout)
        self.select_frame(locator)

    @keyword("Select Nested Frame")
    def select_nested_frame(self, *locators):
        """Selects a frame nested in other frames

        locators: (list) A list of selenium locators(CSS, XPATH, ID, NAME, etc)

        Example (Selects iframe3 after traversing iframe1 and iframe2):
        |  `Select Nested Frame`  |  id-iframe1  |  id=iframe2  |  id=iframe3  |
        """
        for locator in locators:
            self.wait_until_page_contains_element(locator)
            self.select_frame(locator)

    @keyword("Unselect and Select Frame")
    def unselect_and_select_frame(self, locator, timeout=None):
        """This is a series of chained Selenium keywords, that first unselects the current iFrame, then executes a
        wait for and select frame for a new iFrame.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.unselect_frame()
        self.wait_for_and_select_frame(locator, timeout)

    @keyword("Wait For And Select From List")
    def wait_for_and_select_from_list(self, locator, target, timeout=None):
        """This is a series of chained Selenium keywords, that first waits for and focuses on a list element, then
        selects an item(s).\n
        locator:  (string) A selenium locator(CSS, XPATH, ID, NAME, etc)\n
        target: (string, array of strings) The list item(s) to be selected (string, or list of strings)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.select_from_list_by_label(locator, target)

    @keyword("Wait For And Select From List By Value")
    def wait_for_and_select_from_list_by_value(self, locator, target, timeout=None):
        """This is a series of chained Selenium keywords, that first waits for and focuses on a list element, then
        selects an item(s) for it using the value attribute of the item(s) web element.\n
        locator:  (string) A selenium locator(CSS, XPATH, ID, NAME, etc)\n
        target: (string, array of strings) The list item(s) to be selected (string, or list of strings)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.select_from_list_by_value(locator, target)

    @keyword("Wait For And Select From List By Index")
    def wait_for_and_select_from_list_by_index(self, locator, target, timeout=None):
        """This is a series of chained Selenium keywords, that first waits for and focuses on a list element, then
        selects an item(s) for it using the index of the item(s) web element.\n
        locator:  (string) A selenium locator(CSS, XPATH, ID, NAME, etc)\n
        target: (string, array of strings) The list item(s) to be selected (string, or list of strings)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.select_from_list_by_index(locator, target)

    @keyword("Wait For And Mouse Over")
    def wait_for_and_mouse_over(self, locator, timeout=None):
        """This is a series of chained Selenium keywords, that first waits for an element to be visible, then executes a
        mouse over command on it.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.mouse_over(locator)

    @keyword("Wait For And Select Checkbox")
    def wait_for_and_select_checkbox(self, locator, timeout=None):
        """This is a series of chained Selenium keywords, that first waits for an element to be visible, then selects it
        if it's a checkbox.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.select_checkbox(locator)

    @keyword("Wait For And Mouse Over And Click")
    def wait_for_and_mouse_over_and_click(self, locator, timeout=None):
        """This is a series of chained Selenium keywords, that first waits for an element to be visible, executes a
        mouse over command on it, and it finally clicks it.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_for_and_focus_on_element(locator, timeout)
        self.mouse_over(locator)
        self.click_element(locator)

    @keyword("Wait For and Focus On Element")
    def wait_for_and_focus_on_element(self, locator, timeout=None):
        """This is a series of chained Selenium keywords, that first waits for an element to be on the DOM, executes
        Focus on it, then it waits for it to be visible.\n
        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        if not timeout:
            timeout = self.timeout
        self.wait_until_page_contains_element(locator, timeout)
        self.wait_until_element_is_visible(locator, timeout)
        zoomba.wait_until_keyword_succeeds(timeout, 1, "Set Focus To Element", locator)

    @keyword("Wait Until Window Opens")
    def wait_until_window_opens(self, title, timeout=None):
        """Used to get the titles of the current browser windows, then verify that the provided window title
        is among them.\n
        title: (string) The title of the window you are waiting for.\n
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        timeout = time() + float(timeout) if timeout else time() + self.timeout
        while time() < timeout:
            titles = self.get_window_titles()
            if title in titles:
                return
        zoomba.fail("Window with the title: '" + title + "' not found.")

    @keyword("Window Should Not Be Open")
    def window_should_not_be_open(self, title):
        """This is a series of chained Selenium keywords, used to check that a window with the title no longer exists.\n
        title: (string) The title of the window you are expecting to be closed.
        """
        titles = self.get_window_titles()
        zoomba_collections.list_should_not_contain_value(titles, title)

    @keyword("Wait For And Select Window")
    def wait_for_and_select_window(self, title, timeout=None):
        """This is a series of chained Selenium keywords, used to wait until a window with the title argument exists,
        then it selects that window.\n
        title: (string) The title of the window you are waiting for.
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        self.wait_until_window_opens(title, timeout)
        self.switch_window(title)

    @keyword("Scroll To Bottom Of Page")
    def scroll_to_bottom_of_page(self):
        """This keyword scrolls down to the bottom of the page, if jquery is not available it will scroll 20000 pixels

        """
        try:
            height = self.execute_javascript("return window.outerHeight")
        except BaseException as ex:  # lgtm [py/catch-base-exception]
            if ex:
                height = 20000
        self.execute_javascript(f"window.scrollTo(0,{height})")

    @keyword("Wait Until Javascript Is Complete")
    def wait_until_javascript_is_complete(self):
        """This keyword polls the jQuery.active flag, to track execution of AJAX requests. The web application needs
        to have jQuery in order for this to work.

        """
        for each in range(1, 20):
            jquery_started = self.execute_javascript("return jQuery.active==1")
            if jquery_started:
                break
        for _ in range(1, 50):
            jquery_completed = self.execute_javascript("return window.jQuery!=undefined && jQuery.active==0")
            if jquery_completed:
                break
            zoomba.sleep("0.5s")

    @keyword("Get Text From Web Elements List")
    def get_text_from_web_elements_list(self, web_elements_list):
        """This keyword extracts the HTML text of a list of web elements, and returns a list.\n
        web_elements_list: (array of selenium web elements) A selenium web element objects list.\n
        return: (array of strings) A list of strings extracted from the web elements.
        """
        text_element_list = []
        for web_element in web_elements_list:
            element_text = self.get_text(web_element)
            text_element_list.append(element_text)
        return text_element_list

    @keyword("Get Values From Web Elements List")
    def get_values_from_web_elements_list(self, web_elements_list):
        """This keyword extracts the HTML value of a list of web elements, and returns a list.\n
        web_elements_list: (array of selenium web elements) A selenium web element objects list.\n
        return: (array of strings) A list of strings extracted from the web elements.
        """
        value_element_list = []
        for web_element in web_elements_list:
            element_value = self.get_value(web_element)
            value_element_list.append(element_value)
        return value_element_list

    @keyword("Get Vertical Position From Web Elements List")
    def get_vertical_position_from_web_elements_list(self, web_elements_list):
        """This keyword extracts the vertical position in pixels of a list of web elements, and returns a list.\n
        web_elements_list: (array of selenium web elements) A selenium web element objects list.\n
        return: (array of integers) A list of integers of the vertical pixel positions for the provided web elements.
        """
        position_element_list = []
        for web_element in web_elements_list:
            element_position = self.get_vertical_position(web_element)
            position_element_list.append(element_position)
        return position_element_list

    @keyword("Wait Until Window Closes")
    def wait_until_window_closes(self, title, timeout=None):
        """This is a series of chained Selenium keywords, used to wait until a window with the title no longer exists.\n
        title: (string) The title of the window you are expecting to close.
        timeout: (float) Time in seconds to wait, will use global timeout if not set.
        """
        if not timeout:
            timeout = self.timeout
        zoomba.wait_until_keyword_succeeds(timeout, 1, "Window Should Not Be Open", title)

    @keyword("Wait Until Element Contains Value")
    def wait_until_element_contains_value(self, locator, expected_value, timeout=None):
        """Waits until the element ``locator`` appears on the current page and contains value.
        Fails if ``timeout`` expires before the element appears. See
        the `Timeouts` section for more information about using timeouts and
        their default value and the `Locating elements` section for details
        about the locator syntax.
        """
        self.wait_until_page_contains_element(locator, timeout)
        value = self.get_value(locator)
        zoomba.should_contain(value, expected_value)

    @keyword("Create Dictionary From Keys And Values Lists")
    def create_dictionary_from_keys_and_values_lists(self, keys, values):
        """This keyword returns a dictionary, given a list of keys and a list of values of the same length.\n
        keys: (array of strings) A list of keys to be assigned to a dictionary.\n
        values: (array of strings,integers,dictionaries,etc) A list of values to be assigned to a dictionary.\n
        return: (dictionary) A dictionary with the mapped key-value pairs.
        """
        if len(keys) != len(values):
            zoomba.log("The length of the keys and values lists is not the same: \nKeys Length: " +
                       str(len(keys)) + "\nValues Length: " + str(len(values)), "ERROR")
        return dict(zip(keys, values))

    @keyword("Truncate String")
    def truncate_string(self, string, number_of_characters):
        """Truncates a given string to a certain amount of the given number of characters parameter.\n
        string: (string) Original String
        number_of_characters: (integer) Truncation index
        return: (string) A truncated String
        """
        return string[0:number_of_characters]

    @keyword("Save Selenium Screenshot")
    def save_selenium_screenshot(self):
        """Takes a screenshot with a unique filename to be stored in Robot Framework compiled reports.

        If `Set Screenshot Directory` has been set to ``EMBED`` then the screenshot will be embedded into the report
        """
        if is_string(self.screenshot_root_directory):
            if self.screenshot_root_directory.upper() == 'EMBED':
                return self.capture_page_screenshot()
        else:
            timestamp = time()
            filename = 'selenium-screenshot-' + str(timestamp) + '-' + str(next(SCREENSHOT_COUNTER)) + '.png'
            return self.capture_page_screenshot(filename)

    @keyword("Get Element CSS Attribute Value")
    def get_element_css_attribute_value(self, locator, attribute):
        """Get the CSS attribute value of an element. This would be the values within the 'style' tag.

        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)

        attribute: (string) The attribute you wish to get the value for.
        """
        css = self.get_webelement(locator)
        return zoomba.call_method(css, 'value_of_css_property', attribute)

    @keyword("Element CSS Attribute Value Should Be")
    def element_css_attribute_value_should_be(self, locator, attribute, expected_value):
        """Get the CSS attribute value of an element and compare it to an expected value.
        This would be the values within the 'style' tag.

        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)

        attribute: (string) The attribute you wish to get the value for.

        expected_value: (string) The expected attribute value to be compared against the actual
        attribute value
        """
        attribute_value = self.get_element_css_attribute_value(locator, attribute)
        zoomba.should_be_equal(attribute_value, expected_value)

    @keyword("Get React List Labels")
    def get_react_list_labels(self, locator):
        """This keyword grabs the labels from a React-Select list. See https://react-select.com/home for
        more information on React-Select components.

        locator: (string) A selenium locator(CSS, XPATH, ID, NAME, etc)

        """
        react_select_container = self.find_element(locator)
        options = RS.ReactSelect(react_select_container).options()
        return [opt.text for opt in options]



"""
This module is for the ReactSelect class, which mimics the Selenium Library Select class, but for React-Select components.
See: https://github.com/SeleniumHQ/selenium/blob/trunk/py/selenium/webdriver/support/select.py

See the Helpers/ReactHelpersREADME.md for more information on the structure of React-Select components

"""

from selenium.common.exceptions import UnexpectedTagNameException
from selenium.webdriver.common.by import By


class ReactSelect:
    """React Select

    This class is for helper methods for the Zoomba GUI Library, specifically for React-Select components.
    React-Select components use <div> tags instead of the traditional <select> & <option>.
    Zoomba.GUILibrary method Example:
            from ReactSelect import ReactSelect as RS
            react_select_container = self.find_element(locator)
            options = RS.ReactSelect(react_select_container).options()
    """

    def __init__(self, webelement):
        """
        Constructor. A check is made that the given element is, indeed, a DIV tag. If it is not,
        then an UnexpectedTagNameException is thrown.

        :Args:
         - webelement - Web element representing the React-Select container <div> element.

        """
        if webelement.tag_name.lower() != "div":
            raise UnexpectedTagNameException(f"ReactSelect only works on <div> elements, not on <{webelement.tag_name}>")
        self._el = webelement

    def options(self):
        """Returns a list of all options belonging to the React-Select container"""
        self.expand_select_list()
        return self._el.find_elements(By.XPATH, './div[2]/div[1]/div')

    def is_expanded(self):
        """Checks if the React-Select container is expanded by checking if the Menu <div>
        exists as a child of the container
        """
        menu_elements = self._el.find_elements(By.XPATH, './div[2]')
        if len(menu_elements) == 0:
            return False
        if len(menu_elements) == 1:
            return True
        raise LookupError("ReactSelect.is_expanded: Multiple selection menus found")

    def expand_select_list(self):
        """Expands the Select list only if it's collapsed"""
        is_expanded = self.is_expanded()
        if not is_expanded:
            self._el.click()



import json
from robot.libraries.BuiltIn import BuiltIn
from suds.plugin import DocumentPlugin
from suds.client import Client
from suds import WebFault

zoomba = BuiltIn()


class _ObjectNamespacePlugin(DocumentPlugin):
    """_ObjectNamespacePlugin

        This class contains a plugin for use in fixing broken WSDL Namespaces.\n
        defaultNamespace: (string) The default namespace found in the primary WSDL\n


    """
    defaultNamespace = None

    def loaded(self, context):
        """loaded. A plugin run after loading the WSDL but before using it to construct a client.\n
            context: (string) The Raw WSDL and imported schema.\n
        """
        if self.defaultNamespace is None:
            start = context.document.find(b'targetNamespace')
            first_quote = int(context.document.find(b'"', start)) + 1
            second_quote = int(context.document.find(b'"', first_quote))
            self.defaultNamespace = context.document[first_quote:second_quote]
        elif b'tns' not in context.document or b'targetNamespace' not in context.document:
            document_split = context.document.split(b'xmlns', 1)
            context.document = document_split[0]+b'xmlns:tns="'+self.defaultNamespace+b'" targetNamespace="' + self.defaultNamespace + b'" xmlns'+document_split[1]
            number_of_types = context.document.count(b'type="')
            start_location = context.document.find(b'type="')
            for _ in range(number_of_types):
                end_type = context.document.find(b'"', start_location+len(b'type="'))
                if b':' not in context.document[start_location:end_type]:
                    change_type = context.document[start_location:end_type]
                    change_type = change_type.replace(b'"', b'"tns:')
                    context.document = context.document[:start_location] + change_type + context.document[end_type:]
                start_location = context.document.find(b'type="', start_location+1)


class SOAPLibrary(object):
    """Zoomba SOAP Library

        This class is the base Library used to generate automated SOAP Tests in the Zoomba Automation Framework.

    """

    @staticmethod
    def create_soap_session_and_fix_wsdl(host=None, endpoint=None, alias=None, **kwargs):
        """Create Soap Session. This Keyword utilizes the WSDL and directly accesses calls from sudsLibrary.\n
            host: (string) The host url.\n
            endpoint: (string) SOAP API endpoint containing the actions to be referenced.\n
            alias: (string} Sets the alias for the SudsLibrary Framework
            **kwargs: (optional) Parameters that could be included to add options to client creation.
            Current supported parameters are:\n
                set_location: http address\n
        """
        pluginInstance = _ObjectNamespacePlugin()
        plugin_client = Client(host+endpoint + '?WSDL', plugins=[pluginInstance])
        suds_library = BuiltIn().get_library_instance("SudsLibrary")
        suds_library._add_client(plugin_client, alias)
        if 'set_location' in kwargs:
            suds_library.set_location(kwargs['set_location'])

    @staticmethod
    def create_soap_session(host=None, endpoint=None, alias=None, **kwargs):
        """ Create Soap Session. This Keyword utilizes the WSDL to create a soap client.\n
            host: (string) The host url.\n
            endpoint: (string) SOAP API endpoint containing the actions to be referenced.\n
            **kwargs: (optional) Parameters that could be included to add options to client creation.
            Current supported parameters are:\n
                set_location: http address\n
        """
        suds_library = BuiltIn().get_library_instance("SudsLibrary")
        if alias is not None:
            suds_library.create_soap_client(host+endpoint+'?WSDL', alias)
        else:
            suds_library.create_soap_client(host+endpoint+'?WSDL')
        if 'set_location' in kwargs:
            suds_library.set_location(kwargs['set_location'])

    def create_soap_session_and_set_location(self, host=None, endpoint=None, alias=None, set_location=None, fix=False):
        """ Create Soap Session and Set Location. In addition to the client creation, this keyword sets the location
            as specified.\n
            host: (string) The host url.\n
            endpoint: (string) SOAP API endpoint containing the actions to be referenced.\n
            set_location: (string) If set will overwrite the WSDL location with specified address.\n
            If set to None will replace location with host and endpoint specified\n
        """
        if set_location is None:
            set_location = host+endpoint
        if fix:
            self.create_soap_session_and_fix_wsdl(host, endpoint, alias, set_location=set_location)
        else:
            self.create_soap_session(host, endpoint, alias, set_location=set_location)

    @staticmethod
    def call_soap_method_with_list_object(action=None, soap_object=None):
        """ Call Soap Method. Calls soap method with list object \n
            action: (string) SOAP Action to be called.\n
            soap_object: (list) Soap Object in list format, list must be ordered wrt schema\n
        """
        suds_library = BuiltIn().get_library_instance("SudsLibrary")
        return suds_library.call_soap_method(action, *soap_object)

    @staticmethod
    def call_soap_method_with_object(action=None, **soap_object):
        """ Call Soap Method with dictionary object. Calls soap method \n
            action: (string) SOAP Action to be called.\n
            soap_object: (dict) Soap Object in dict format, dict must contain all required parts of schema object. \n
        """
        suds_library = BuiltIn().get_library_instance("SudsLibrary")
        client = suds_library._client()
        method = getattr(client.service, action)
        try:
            received = method(**soap_object)
        except WebFault as e:
            received = e.fault
        return received

    @staticmethod
    def create_wsdl_objects(wsdl_type=None, object_dict=None):
        """ Create Wsdl Objects. This Keyword utilizes the WSDL to create a WSDL object based on the information
            provided.\n
            wsdl_type: (string) Wsdl object to be created.\n
            object_dict: (dict) Python Dictionary containing values and nested dictionaries with construction similar to
            wsdl defined objects.\n
            return: (response object) Returns the SOAP client object.\n
        """
        client = BuiltIn().get_library_instance("SudsLibrary")._client()
        request_object = client.factory.create(wsdl_type)
        _build_wsdl_objects(client, request_object, object_dict)
        return request_object

    @staticmethod
    def convert_soap_response_to_json(soap_response=None):
        """ Convert Soap Response To Dictionary: This keyword builds a dictionary from the sudsLibrary response\n
            json_actual_response: (request response object) The response from an API.\n
            return: There is no actual returned output, other than error messages when comparisons fail.\n
        """
        a = _build_dict_from_response(soap_response)
        return json.dumps(a)


def _build_dict_from_response(soap_response=None):
    """
    Build Dict From Response: This keyword builds a dictionary from the sudsLibrary response.\n
    :param soap_response: sudsLibrary response.\n
    :return: python dictionary.\n
    """
    try:
        response_dictionary = dict(soap_response)
    except:  # lgtm [py/catch-base-exception]
        zoomba.log(message='Argument Passed Was Not Iterable', level='INFO')
        return soap_response
    new_response = {}

    for _ in range(len(response_dictionary)):
        key, value = response_dictionary.popitem()
        if isinstance(value, list):
            temp_list = []
            for item in value:
                if str(type(item)) == "<type 'instance'>" or 'sudsobject' in str(type(item)):
                    temp_item = _build_dict_from_response(item)
                    temp_list.append(temp_item)
                else:
                    temp_list.append(str(item))
            new_response[key] = temp_list
        elif str(type(value)) == "<type 'instance'>" or 'sudsobject' in str(type(value)):
            temp_dict = _build_dict_from_response(value)
            new_response[key] = temp_dict
        else:
            new_response[key] = str(value)
    return new_response


def _build_wsdl_objects(client=None, request_object=None, object_dict=None):
    """ Build Wsdl Objects. This Keyword utilizes the WSDL to build a wsdl object in a recursive manner.\n
        client: (SudsLibrary._client() instance) The current client session.\n
        request_object: (SudsLibrary._client.factory() instance) Wsdl object to be built.\n
        object_dict: (dict) Dictionary containing data to create request_object.\n

    """
    for key, value in object_dict.items():
        if isinstance(value, dict):
            try:
                temp_object = _wsdl_sub_builder(client, value)
                request_object.__setattr__(key, temp_object)
            except BaseException as ex:   #lgtm [py/catch-base-exception]
                if ex:
                    zoomba.log('Failed to define wsdl_object_type for child object. [' + key + ']', level='ERROR')
        elif isinstance(value, list):
            temp_list = []
            for item in value:
                if isinstance(item, dict):
                    temp_object = _wsdl_sub_builder(client, item)
                    temp_list.append(temp_object)
                else:
                    temp_list.append(item)
            request_object.__setattr__(key, temp_list)
        else:
            request_object.__setattr__(key, value)


def _wsdl_sub_builder(client=None, object_dict=None):
    """
    Wsdl Sub Builder. This Keyword creates children objects for wsdl objects.\n
    client: (SudsLibrary._client() instance) The current client session.\n
    object_dict:  (dict) Dictionary containing data to create request_object.\n
    (method only) wsdl_object_type: Wsdl object Type of wsdl sub-object.\n
    return: returns child object\n
    """
    sub_type = object_dict.pop('wsdl_object_type')
    temp_object = client.factory.create(sub_type)
    _build_wsdl_objects(client, temp_object, object_dict)
    return temp_object



import datetime
import json

from RequestsLibrary import RequestsLibrary, utils
from dateutil.parser import parse
from urllib3.exceptions import InsecureRequestWarning
from requests.packages import urllib3
from robot.libraries.BuiltIn import BuiltIn
from robot.utils.dotdict import DotDict

zoomba = BuiltIn()
requests_lib = RequestsLibrary()


class APILibrary:
    """Zoomba API Library

        This class is the base Library used to generate automated API Tests in the Robot Automation Framework.
        It has been generated to accommodate the RESTful API design pattern.
    """
    def __init__(self):
        self.suppress_warnings = False

    def suppress_insecure_request_warnings(self, suppress="True"):
        """Suppress Insecure Request Warnings. This keyword suppresses or un-suppresses insecure request warnings\n
        suppress: (bool) True/False to suppress warnings, default is True\n
        Examples:
        | Suppress Insecure Request Warnings                       #Suppresses Insecure Request Warnings
        | Suppress Insecure Request Warnings | suppress=True |     #Suppresses Insecure Request Warnings
        | Suppress Insecure Request Warnings | False |    #Does Not Suppresses Insecure Request Warnings
        """
        self.suppress_warnings = "FALSE" not in suppress.upper()

    def call_get_request(self, headers=None, endpoint=None, fullstring=None, cookies=None, timeout=None):
        """ Generate a GET Request. This Keyword is basically a wrapper for get_request from the RequestsLibrary.\n
            headers: (dictionary) The headers to be sent as part of the request.\n
            endpoint: (string) The string that identifies the url endpoint of the App that receives API requests.\n
            fullstring: (string) A string that contains the rest of the url that identifies a specific API/Webservice\n
            timeout: (float) Time in seconds for the api to respond\n
            return: (response object) Returns the request response object, which includes headers, content, etc.
            along with any query parameters.
        """
        if self.suppress_warnings:
            urllib3.disable_warnings(InsecureRequestWarning)
        requests_lib.create_session("getapi", endpoint, headers, cookies=cookies, timeout=timeout)
        resp = requests_lib.get_on_session("getapi", fullstring, timeout=timeout, expected_status='any')
        return _convert_resp_to_dict(resp)

    def call_post_request(self, headers=None, endpoint=None, fullstring=None, data=None, files=None, cookies=None, timeout=None):
        """ Generate a POST Request. This Keyword is basically a wrapper for post_request from the RequestsLibrary.\n
            headers: (dictionary) The headers to be sent as part of the request.\n
            endpoint: (string) The string that identifies the url endpoint of the App that receives API requests.\n
            fullstring: (string) A string that contains the rest of the url that identifies a specific API/Webservice
            along with any query parameters.\n
            timeout: (float) Time in seconds for the api to respond\n
            data: (json) The JSON object to be sent on the body of the request to be used by the specific Web service.\n
            files: (json) A JSON object that sends in the body of the request to be used by the specific Web service.\n
            return: (response object) Returns the request response object, which includes headers, content, etc.
        """
        if self.suppress_warnings:
            urllib3.disable_warnings(InsecureRequestWarning)
        session = requests_lib.create_session("postapi", endpoint, headers, cookies=cookies, timeout=timeout)
        data = utils.format_data_according_to_header(session, data, headers)
        resp = requests_lib.post_on_session("postapi", fullstring, data, files=files, timeout=timeout, expected_status='any')
        return _convert_resp_to_dict(resp)

    def call_delete_request(self, headers=None, endpoint=None, fullstring=None, cookies=None, timeout=None):
        """ Generate a DELETE Request. This Keyword is basically a wrapper for delete_request from the RequestsLibrary.\n
            headers: (dictionary) The headers to be sent as part of the request.\n
            endpoint: (string) The string that identifies the url endpoint of the App that receives API requests.\n
            fullstring: (string) A string that contains the rest of the url that identifies a specific API/Webservice
            along with any query parameters.\n
            timeout: (float) Time in seconds for the api to respond\n
            return: (response object) Returns the request response object, which includes headers, content, etc.
        """
        if self.suppress_warnings:
            urllib3.disable_warnings(InsecureRequestWarning)
        requests_lib.create_session("deleteapi", endpoint, headers, cookies=cookies, timeout=timeout)
        resp = requests_lib.delete_on_session("deleteapi", fullstring, timeout=timeout, expected_status='any')
        return _convert_resp_to_dict(resp)

    def call_patch_request(self, headers=None, endpoint=None, fullstring=None, data=None, cookies=None, timeout=None):
        """ Generate a PATCH Request. This Keyword is basically a wrapper for patch_request from the RequestsLibrary.\n
            headers: (dictionary) The headers to be sent as part of the request.\n
            endpoint: (string) The string that identifies the url endpoint of the App that receives API requests.\n
            fullstring: (string) A string that contains the rest of the url that identifies a specific API/Webservice
            along with any query parameters.\n
            timeout: (float) Time in seconds for the api to respond\n
            data: (json) The JSON object to be sent on the body of the request to be used by the specific Web service.\n
            return: (response object) Returns the request response object, which includes headers, content, etc.
        """
        if self.suppress_warnings:
            urllib3.disable_warnings(InsecureRequestWarning)
        session = requests_lib.create_session("patchapi", endpoint, headers, cookies=cookies, timeout=timeout)
        data = utils.format_data_according_to_header(session, data, headers)
        resp = requests_lib.patch_on_session("patchapi", fullstring, data, timeout=timeout, expected_status='any')
        return _convert_resp_to_dict(resp)

    def call_put_request(self, headers=None, endpoint=None, fullstring=None, data=None, cookies=None, timeout=None):
        """ Generate a PUT Request. This Keyword is basically a wrapper for put_request from the RequestsLibrary.\n
            headers: (dictionary) The headers to be sent as part of the request.\n
            endpoint: (string) The string that identifies the url endpoint of the App that receives API requests.\n
            fullstring: (string) A string that contains the rest of the url that identifies a specific API/Webservice
            along with any query parameters.\n
            timeout: (float) Time in seconds for the api to respond\n
            data: (json) The JSON object to be sent on the body of the request to be used by the specific Web service.\n
            return: (response object) Returns the request response object, which includes headers, content, etc.
        """
        if self.suppress_warnings:
            urllib3.disable_warnings(InsecureRequestWarning)
        session = requests_lib.create_session("putapi", endpoint, headers, cookies=cookies, timeout=timeout)
        data = utils.format_data_according_to_header(session, data, headers)
        resp = requests_lib.put_on_session("putapi", fullstring, data, timeout=timeout, expected_status='any')
        return _convert_resp_to_dict(resp)

    def create_connection(self, endpoint, method, data, headers=None, cookies=None, timeout=None):
        """*DEPRECATED!!* Use 'Call Post Request' instead."""
        return self.call_post_request(headers=headers, endpoint=endpoint, fullstring=method, data=data, cookies=cookies, timeout=timeout)

    def validate_response_contains_expected_response(self, json_actual_response, expected_response_dict,
                                                     ignored_keys=None, full_list_validation=False, identity_key="id",
                                                     sort_lists=False, **kwargs):
        """ This is the most used method for validating Request responses from an API against a supplied
            expected response. It performs an object to object comparison between two json objects, and if that fails,
            a more in depth method is called to find the exact discrepancies between the values of the provided objects.
            Additionally, a list of keys to ignore on the comparison may be supplied, for keys' values to be ignored./n

            json_actual_response: (request response object) The response from an API.\n
            expected_response_dict: (json) The expected response, in json format.\n
            ignored_keys: (strings list) A list of strings of the keys to be ignored on the validation.\n
            full_list_validation: (bool) Check that the entire list matches the expected response, defaults to False.\n
            identity_key: (string) Key to match items to, defaults to 'id'.\n
            sort_lists: (bool) Sort lists before doing key by key validation, defaults to False.\n
            **kwargs: (dict) Currently supported kwargs are margin_type and margin_amt\n
            margin_type: (string) The type of unit of time to be used to generate a delta for the date comparisons.\n
            margin_amt: (string/#) The amount of units specified in margin_type to allot for difference between dates.\n
            return: There is no actual returned output, other than error messages when comparisons fail.\n
        """
        if not json_actual_response:
            zoomba.fail("The Actual Response is Empty.")
            return
        actual_response_dict = json.loads(json_actual_response)
        unmatched_keys_list = []
        if not isinstance(actual_response_dict, list) and actual_response_dict:
            if actual_response_dict == expected_response_dict:
                return
            self.key_by_key_validator(actual_response_dict, expected_response_dict, ignored_keys,
                                      unmatched_keys_list, full_list_validation=full_list_validation, sort_lists=sort_lists, **kwargs)
            self.generate_unmatched_keys_error_message(unmatched_keys_list)
            return
        if isinstance(actual_response_dict, list) and actual_response_dict:
            if full_list_validation:
                return self.full_list_validation(actual_response_dict, expected_response_dict, unmatched_keys_list,
                                                 ignored_keys, sort_lists=sort_lists, **kwargs)
            for exp_item in expected_response_dict:
                for actual_item in actual_response_dict:
                    try:
                        if exp_item[identity_key] == actual_item[identity_key]:
                            self.key_by_key_validator(actual_item, exp_item, ignored_keys, unmatched_keys_list,
                                                      full_list_validation=full_list_validation, sort_lists=sort_lists, **kwargs)
                            self.generate_unmatched_keys_error_message(unmatched_keys_list)
                            break
                        elif actual_response_dict[-1] == actual_item:
                            zoomba.fail('Item was not within the response:\n' + str(exp_item))
                            return
                        else:
                            continue
                    except KeyError:
                        zoomba.fail('KeyError: "' + identity_key + '" Key was not in the response')
                        break
        else:
            zoomba.fail("The Actual Response is Empty.")

    def validate_response_contains_expected_response_only_keys_listed(self, json_actual_response, expected_response,
                                                                      key_list):
        """ This keyword is used for validating that a specific set of key-value pairs are contained on Request
            responses from an API.\n
            json_actual_response: (request response object) The response from an API.\n
            jsonExpectedResponse: (json) The expected response, in json format.\n
            key_list: (strings list) A list of strings of the keys to be included on the validation.\n
            return: There is no actual returned output, other than error messages when comparisons fail.\n
        """
        actual_response_dict = json.loads(json_actual_response)
        if not isinstance(actual_response_dict, list):
            for expected_key in key_list:
                if expected_key not in actual_response_dict:
                    zoomba.fail("The response does not contain the key '" + expected_key + "'")
                    continue
                if actual_response_dict[expected_key] != expected_response[expected_key]:
                    zoomba.fail("The value for the key '" + expected_key + "' doesn't match the response:" +
                                "\nExpected: " + expected_response[expected_key] +
                                "\nActual: " + actual_response_dict[expected_key])
            return
        for expected_key in key_list:
            if expected_key not in actual_response_dict[0]:
                zoomba.fail("The response does not contain the key '" + expected_key + "'")
                continue
            if actual_response_dict[0][expected_key] != expected_response[0][expected_key]:
                zoomba.fail("The value for the key '" + expected_key + "' doesn't match the response:" +
                            "\nExpected: " + expected_response[0][expected_key] +
                            "\nActual: " + actual_response_dict[0][expected_key])
        return

    def validate_response_contains_correct_number_of_items(self, json_actual_response, number_of_items):
        """ This keyword is used to validate the number of returned items on Request responses from an API.\n
            json_actual_response: (request response object) The response from an API.\n
            number_of_items: (integer) The expected number of items.\n
            return: There is no actual returned output, other than error messages when comparisons fail.\n
        """
        actual_response_dict = json.loads(json_actual_response)
        if isinstance(number_of_items, str):
            number_of_items = number_of_items.upper()
            if number_of_items == "IGNORE":
                return True
        elif not isinstance(number_of_items, int):
            zoomba.fail("Did not pass number or string value, function expects a number or string 'IGNORE'.")
            return

        if isinstance(actual_response_dict, list):
            if len(actual_response_dict) != int(number_of_items):
                zoomba.fail('API is returning ' + str(
                    len(actual_response_dict)) + ' instead of the expected ' + str(number_of_items) + ' result(s).')
        else:
            zoomba.fail("The response is not a list:\nActual Response:\n" + str(actual_response_dict))

    def key_by_key_validator(self, actual_dictionary, expected_dictionary, ignored_keys=None, unmatched_keys_list=None,
                             parent_key=None, full_list_validation=False, sort_lists=False, **kwargs):
        """ This method is used to find and verify the value of every key in the expectedItem dictionary when compared
            against a single dictionary actual_item, unless any keys are included on the ignored_keys array./n

            actual_item: (array of dictionaries) The list of dictionary items extracted from a json Response.\n
            ExpectedItem: (dictionary) The expected item with the key to be validated.\n
            ignored_keys: (strings list) A list of strings of the keys to be ignored on the validation.\n
            full_list_validation: (bool) Check that the entire list matches the expected response, defaults to False.\n
            sort_lists: (bool) Sort lists before doing key by key validation, defaults to False.\n
            **kwargs: (dict) Currently supported kwargs are margin_type and margin_amt\n
            margin_type: (string) The type of unit of time to be used to generate a delta for the date comparisons.\n
            margin_amt: (string/#) The amount of units specified in margin_type to allot for difference between dates.\n
            return: (boolean) If the method completes successfully, it returns True. Appropriate error messages are
            returned otherwise.\n
        """
        if len(actual_dictionary) != len(expected_dictionary):
            zoomba.fail("Collections not the same length:"
                        "\nActual length: " + str(len(actual_dictionary)) +
                        "\nExpected length " + str(len(expected_dictionary)))
            return
        for key, value in expected_dictionary.items():
            if ignored_keys and key in ignored_keys:
                continue
            if key not in actual_dictionary:
                zoomba.fail("Key not found in Actual : " + str(actual_dictionary) + " Key: " + str(key))
                continue
            if isinstance(value, list):
                if full_list_validation and len(value) != len(actual_dictionary[key]):
                    zoomba.fail("Arrays not the same length:" +
                                "\nExpected: " + str(value) +
                                "\nActual: " + str(actual_dictionary[key]))
                    continue
                self._key_by_key_list(key, value, actual_dictionary, unmatched_keys_list, ignored_keys, parent_key,
                                      full_list_validation=full_list_validation, sort_lists=sort_lists, **kwargs)
            elif isinstance(value, dict):
                self._key_by_key_dict(key, value, actual_dictionary, expected_dictionary, unmatched_keys_list,
                                      ignored_keys, full_list_validation=full_list_validation, sort_lists=sort_lists,
                                      **kwargs)
            elif isinstance(expected_dictionary[key], str) and not expected_dictionary[key].isdigit():
                try:
                    parse(expected_dictionary[key])
                    self.date_string_comparator(value, actual_dictionary[key], key, unmatched_keys_list, **kwargs)
                except (ValueError, TypeError):
                    if value == actual_dictionary[key]:
                        continue
                    else:
                        unmatched_keys_list.append(("------------------\n" + "Key: " + str(key),
                                                    "Expected: " + str(value),
                                                    "Actual: " + str(actual_dictionary[key])))
            elif value == actual_dictionary[key]:
                continue
            else:
                unmatched_keys_list.append(("------------------\n" + "Key: " + str(key), "Expected: " + str(value),
                                            "Actual: " + str(actual_dictionary[key])))
        return True

    def date_string_comparator(self, expected_date, actual_date, key, unmatched_keys_list, **kwargs):
        """This Method is used to validate a single property on a JSON object of the Date Type.
        It Validates for any the following Date Formats:
        %Y-%m-%dT%H:%M:%S, %Y-%m-%dT%H:%M:%SZ, %Y-%m-%dT%H:%M:%S.%f, %Y-%m-%dT%H:%M:%S.%fZ

        expected_date: (string) The Expected date string the key being validated.\n
        actual_date: (string) The Actual date string of the key being validated.\n
        key: (string) The key being validated.\n
        unmatched_keys_list (list): List of keys that are unvalidated - to be passed to error handling method.
        **kwargs: (dict) Currently supported kwargs are margin_type and margin_amt\n
        margin_type: (string) The type of unit of time to be used to generate a delta for the date comparisons.\n
        margin_amt: (string/#) The amount of units specified in margin_type to allot for difference between dates.\n
        """
        if expected_date == actual_date:
            return
        expected_utc = _date_format(expected_date, key, unmatched_keys_list, "Expected")
        actual_utc = _date_format(actual_date, key, unmatched_keys_list, "Actual")
        if expected_utc and actual_utc:
            self.date_comparator(expected_utc, actual_utc, key, unmatched_keys_list, **kwargs)

    def date_comparator(self, expected_date, actual_date, key, unmatched_keys_list, margin_type="minutes",
                        margin_amt=10):
        """This method compares two date values, given a certain margin type(minutes, seconds, etc),
        and a margin amount (int). If the two dates are not within the margin amount for the margin type, I.E. within
        10 minutes of difference, it asserts False, and returns an error message.

        expected_date: (date) The Expected date value of the key being validated.\n
        actual_date: (date) The Actual date value of the key being validated.\n
        key: (string) The key being validated.\n
        unmatched_keys_list: (list) List of Date keys that are not within the accepted margin_type
        and margin_amt resolution\n
        margin_type: (string) The type of unit of time to be used to generate a delta for the date comparisons.\n
        margin_amt: (integer) The amount of units specified in margin_type to allot for difference between dates.\n
        """
        arg_dict = {margin_type: int(margin_amt)}
        margin = datetime.timedelta(**arg_dict)
        if expected_date - margin <= actual_date <= expected_date + margin:
            return
        unmatched_keys_list.append(("------------------\n" + "Dates Not Close Enough\nKey: " + str(key),
                                    "Expected: " + str(expected_date),
                                    "Actual: " + str(actual_date)))

    def generate_unmatched_keys_error_message(self, unmatched_keys):
        """ This method is only used as an internal call from other validating methods to generate an error string
            containing every unmatched key when a validation fails.\n
            unmatchedKeys: (array of key/value pairs) An array containing the unmatched keys during a validation.\n
        """
        if unmatched_keys:
            keys_error_msg = "Key(s) Did Not Match:\n"
            for key_error_tuple in unmatched_keys:
                for key_error in key_error_tuple:
                    keys_error_msg += str(key_error) + "\n"
            zoomba.fail(keys_error_msg + "\nPlease see differing value(s)")

    def _key_by_key_list(self, key, value, actual_dictionary, unmatched_keys_list=None, ignored_keys=None,
                         parent_key=None, full_list_validation=False, sort_lists=False, **kwargs):
        if sort_lists and isinstance(value, list):
            try:
                value = list(map(dict, sorted(list(i.items()) for i in value)))
            except AttributeError:
                pass
        for index, item in enumerate(value):
            if isinstance(item, str):
                if value != actual_dictionary[key]:
                    if sort_lists:
                        if sorted(value) != sorted(actual_dictionary[key]):
                            zoomba.fail("Arrays do not match:" +
                                        "\nExpected: " + str(sorted(value)) +
                                        "\nActual: " + str(sorted(actual_dictionary[key])))
                            continue
                    else:
                        zoomba.fail("Arrays do not match:" + \
                                    "\nExpected: " + str(value) + \
                                    "\nActual: " + str(actual_dictionary[key]) + \
                                    "\nIf this is simply out of order try 'sort_list=True'")
                        continue
            else:
                if len(actual_dictionary[key]) == 0:
                    actual_item = ''
                else:
                    if sort_lists:
                        actual_dictionary[key] = list(map(dict, sorted(list(i.items()) for i in actual_dictionary[key])))
                    actual_item = actual_dictionary[key][index]
                temp_actual_dict = {key: actual_item}
                temp_expected_dict = {key: item}
                if unmatched_keys_list:
                    current_unmatched_length = len(unmatched_keys_list)
                else:
                    current_unmatched_length = 0
                self.key_by_key_validator(temp_actual_dict, temp_expected_dict,
                                          ignored_keys, unmatched_keys_list, parent_key=key,
                                          full_list_validation=full_list_validation, sort_lists=sort_lists, **kwargs)
                if unmatched_keys_list is None:
                    continue
                else:
                    _unmatched_list_check(unmatched_keys_list, current_unmatched_length,
                                          key, index, parent_key, is_list=True)

    def _key_by_key_dict(self, key, value, actual_dictionary, expected_dictionary, unmatched_keys_list=None,
                         ignored_keys=None, full_list_validation=False, sort_lists=False, **kwargs):
        try:
            if len(value) != len(actual_dictionary[key]):
                zoomba.fail("Dicts do not match:" +
                            "\nExpected: " + str(value) +
                            "\nActual: " + str(actual_dictionary[key]))
                return
        except TypeError:
            zoomba.fail("Dicts do not match:" +
                        "\nExpected: " + str(value) +
                        "\nActual is not a valid dictionary.")
            return
        if unmatched_keys_list is not None:
            current_unmatched_length = len(unmatched_keys_list)
        self.key_by_key_validator(actual_dictionary[key], expected_dictionary[key],
                                  ignored_keys, unmatched_keys_list, parent_key=key,
                                  full_list_validation=full_list_validation, sort_lists=sort_lists, **kwargs)
        if unmatched_keys_list is None:
            return
        _unmatched_list_check(unmatched_keys_list, current_unmatched_length, key)

    def full_list_validation(self, actual_response_dict, expected_response_dict, unmatched_keys_list, ignored_keys=None,
                             sort_lists=False, **kwargs):
        if actual_response_dict == expected_response_dict:
            return
        for actual_item, expected_item in zip(actual_response_dict, expected_response_dict):
            self.key_by_key_validator(actual_item, expected_item, ignored_keys, unmatched_keys_list,
                                      full_list_validation=True, sort_lists=sort_lists, **kwargs)
        if unmatched_keys_list:
            unmatched_keys_list.append(("------------------\n" + "Full List Breakdown:",
                                        "Expected: " + str(expected_response_dict),
                                        "Actual: " + str(actual_response_dict)))
            self.generate_unmatched_keys_error_message(unmatched_keys_list)
        return


def _unmatched_list_check(unmatched_keys_list, current_unmatched_length, key, index=None, parent_key=None,
                          is_list=False):
    if len(unmatched_keys_list) > current_unmatched_length and parent_key == key:
        for new_index in range(len(unmatched_keys_list) - current_unmatched_length):
            reverse_index = -1 * (new_index + 1)
            unmatched_tuple = unmatched_keys_list[reverse_index]
            split_key_string = unmatched_tuple[0].split("Key: " + parent_key)
            new_key_string = split_key_string[0] + "Key: " + parent_key + "[" + str(index) + "]" + split_key_string[1]
            unmatched_keys_list[reverse_index] = (new_key_string, *unmatched_tuple[1:])
    elif len(unmatched_keys_list) > current_unmatched_length and parent_key is not None:
        for new_index in range(len(unmatched_keys_list) - current_unmatched_length):
            reverse_index = -1 * (new_index + 1)
            unmatched_tuple = unmatched_keys_list[reverse_index]
            if "Key: " + str(key) not in unmatched_tuple[0]:
                split_key_string = unmatched_tuple[0].split("Key: ")
                if is_list:
                    new_key_string = split_key_string[0] + "Key: " + key + "[" + str(index) + "]." + split_key_string[1]
                else:
                    new_key_string = split_key_string[0] + "Key: " + key + "." + split_key_string[1]
                unmatched_keys_list[reverse_index] = (new_key_string, *unmatched_tuple[1:])
    elif len(unmatched_keys_list) > current_unmatched_length and is_list:
        for new_index in range(len(unmatched_keys_list) - current_unmatched_length):
            reverse_index = -1 * (new_index + 1)
            unmatched_tuple = unmatched_keys_list[reverse_index]
            if str(key) + "[" + str(index) + "]" not in unmatched_tuple[0]:
                split_key_string = unmatched_tuple[0].split("Key: ")
                if key == split_key_string[1]:
                    new_key_string = split_key_string[0] + "Key: " + key + "[" + str(index) + "]"
                else:
                    new_key_string = split_key_string[0] + "Key: " + key + "[" + str(index) + "]" + \
                                     "." + split_key_string[1]
                unmatched_keys_list[reverse_index] = (new_key_string, *unmatched_tuple[1:])


def _date_format(date_string, key, unmatched_keys_list, date_type, date_format=None):
    formatted_date = None
    if date_format is None:
        try:
            formatted_date = datetime.datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S')
        except ValueError:
            try:
                formatted_date = datetime.datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%SZ')
            except ValueError:
                try:
                    formatted_date = datetime.datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%f')
                except ValueError:
                    try:
                        formatted_date = datetime.datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%fZ')
                    except ValueError:
                        try:
                            formatted_date = parse(date_string, fuzzy=True)
                            formatted_date = str(formatted_date).replace('+00:00', 'Z')
                            formatted_date = formatted_date.replace(' ', 'T')
                            formatted_date = datetime.datetime.strptime(formatted_date, "%Y-%m-%dT%H:%M:%S.%fZ")
                        except ValueError:
                            unmatched_keys_list.append(("------------------\nKey: " + str(key),
                                                        date_type + " Date Not Correct Format:",
                                                        "Expected Formats: %Y-%m-%dT%H:%M:%S",
                                                        "                  %Y-%m-%dT%H:%M:%SZ",
                                                        "                  %Y-%m-%dT%H:%M:%S.%f",
                                                        "                  %Y-%m-%dT%H:%M:%S.%fZ",
                                                        "Date: " + str(date_string)))

    else:
        try:
            formatted_date = datetime.datetime.strptime(date_string, date_format)
        except ValueError:
            unmatched_keys_list.append(("------------------\nKey: " + str(key),
                                        date_type + " Date Not Correct Format:",
                                        "Expected Format: " + date_format,
                                        "Date: " + str(date_string)))
    return formatted_date


def _convert_resp_to_dict(response):
    new_response = {
        item: getattr(response, item)
        for item in dir(response)
        if item[0] != '_'
    }

    return DotDict(new_response)



#!/usr/bin/env python3

import version

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open("docs/README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(name='robotframework-zoomba',
      version=version.VERSION,
      description='Robot Framework mini-framework.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/Accruent/zoomba',
      maintainer='Alex Calandra, Michael Hintz, Keith Smoland, Matthew Giardina, Brandon Wolfe, Neil Howell,'
                 ' Tommy Hoang',
      maintainer_email='robosquad@accruent.com',
      license='GPL-3.0',
      keywords='Robot Framework robot-framework selenium requests appium soap winappdriver appium robotframework'
               'desktop windows zoomba python robotframework-library appium-windows appiumlibrary api-rest api '
               'soap-api appium-mobile mobile',
      platforms='any',
      install_requires=requirements,
      extras_require={
        'testing': [
          'Appium-Python-Client',
          'mock'
        ]
      },
      classifiers="""
        Development Status :: 5 - Production/Stable
        Operating System :: OS Independent
        Programming Language :: Python :: 3
        Topic :: Software Development :: Testing
        Framework :: Robot Framework :: Library
        """.strip().splitlines(),
      package_dir={'': 'src'},
      packages=['Zoomba', 'Zoomba/Helpers']
      )
